/*****************************************************************************
*  dl_vectorlabels.c -- vector label routines
*
*  Calling these routines results in calls to dl_moveto() and dl_drawto().
*----------------------------------------------------------------------------
*  Externally visible functions are:
*
*     DL_RECT dl_labelbounds(DL_LABEL *label)
*                Return bounding box for given label.
*
*     void    dl_drawlabel(DL_LABEL *label)
*                Draw label with given attributes.
*
*     void    dl_fitlabel(DL_LABEL *label, DL_RECT *box)
*                Draw label to fill current drawing library window.
*
*****************************************************************************/

#include <windows.h>
#include <ctype.h>
#include <math.h>
#include "dl_defs.h"

#undef DL_RADIAN
#define DL_RADIAN(x) (DL_PI*(x)/180.0f)

#define DM_ALPHACOUNT    128     /* number of chars in font */
#define DM_ALPHALENGTH    80     /* maximum length of char description */

static int    fontbuilt;		/* Font table built yet? */
static char  *font[DM_ALPHACOUNT];      /* string char descriptions */
static float  wid[DM_ALPHACOUNT];       /* char widths */
static float  up[DM_ALPHACOUNT];        /* char ascents */
static float  down[DM_ALPHACOUNT];      /* char descents */

static DL_RECT	vectorchar(unsigned char c, char *vstr);
static void	font_init();
static void	cmove(DL_POINT loc, float size, float angle, char cx, char cy);
static void	cdraw(DL_POINT loc, float size, float angle, int cx, int cy);
static DL_POINT drawchar(DL_POINT loc, float size, float angle, char c);
static dt_Real	charvalue(char c);
static DL_POINT rotatepoint(DL_POINT point, float theta);
static DL_POINT shiftpoint(DL_POINT point, DL_POINT offset, float angle);

/*===========================================================================
 *   dl_fitlabel() -- Draw label to fill current drawing library window.
 */

void DX_API_PUBLIC dl_fitlabel(DL_LABEL *lab, DL_RECT *box)
{
   DL_LABEL label;
   DL_RECT  bb;
   float    height, width, room;

   /*** Measure required width and height ***/

   width  = box->xmax - box->xmin;
   height = box->ymax - box->ymin;

   /*** Measure label when fit vertically ***/

   label.text  = lab->text;
   label.sizes = lab->sizes;
   label.super = lab->super;
   label.loc.x = box->xmin + 0.5f * width;
   label.loc.y = box->ymin + 0.5f * height;
   label.size  = 0.8f * height;
   label.angle = 0.0f;
   strcpy(label.tag, "MC");

   bb = dl_labelbounds(&label);

   /*** Shrink to fit horizontally if needed. (ASPECT?) ***/

   room = 0.9f * width; /* / dl_aspect(); */
   if (bb.xmax > room) label.size *= room / bb.xmax;

   /* if (bb->xmax > 0.9 * width) label.size *= 0.9 * width / bb->xmax; */

   /*** Draw the label. ***/

   dl_drawlabel(&label);
}

/*===========================================================================
 *  dl_drawlabel() -- draw label with given attributes
 */

void DX_API_PUBLIC dl_drawlabel(DL_LABEL *lab)
{

   DL_RECT    box;
   DL_POINT   p, rp, sp, loc, shift;
   float      siznow, offset;
   char       *text, *sizes, *super;

   /*** We should collect together dl_drawlabel() and dl_labelbounds(). ***/

   static float rsizes[10] = {
	1.0f, 0.9f, 0.8f, 0.7f, 0.6f, 0.5f, 0.4f, 0.3f, 0.2f, 0.1f
   };
   static float rsuper[10] = {
	0.0f, -0.8f, -0.6f, -0.4f, -0.2f, 0.0f, 0.2f, 0.4f, 0.6f, 0.8f
   };
    
   shift.x = 0;
   shift.y = 0;
   loc     = lab->loc;

   /*** If tag is not "LL", we must make some adjustments. ***/

   if (lab->tag[0] != 'L' || lab->tag[1] != 'L') {
      box = dl_labelbounds(lab);
      if (lab->tag[0] == 'C')
         /* shift.y = -0.5 * DL_MIN(box.ymax, 0.8 * lab->size); */
         shift.y = -0.4f * lab->size;
      else if (lab->tag[0] == 'M')
         shift.y = -0.5f * (box.ymax + box.ymin);
      else if (lab->tag[0] == 'U')
         /* shift.y = - DL_MIN(box.ymax, 0.8 * lab->size); */
         shift.y = -0.8f * lab->size;
      else if (lab->tag[0] == 'T')
         shift.y = - box.ymax;
      else if (lab->tag[0] == 'B')
         shift.y = - box.ymin;
      else
         shift.y = 0;

      if (lab->tag[1] == 'C')
         shift.x = -0.5f * box.xmax;
      else if (lab->tag[1] == 'R')
         shift.x = - box.xmax;
      else
         shift.x = 0;
   }

   p.x = shift.x;
   p.y = shift.y;

   /*** Compute the "shifted" point. ***/

   loc    = shiftpoint(loc, p, -lab->angle);

   offset = 0;
   siznow = lab->size;
   p.x    = 0;

   text   = lab->text;
   sizes  = lab->sizes;
   super  = lab->super;
   while (*text) {
      if (*sizes) {
         if (isdigit(*sizes)) siznow = lab->size * rsizes[*sizes - '0'];
         sizes++;
      }
      if (*super) {
         if (isdigit(*super)) offset = lab->size * rsuper[*super - '0'];
         super++;
      }

      p.y = offset;
      rp  = rotatepoint(p, DL_RADIAN(-lab->angle));
      loc = shiftpoint(loc, p, -lab->angle);

      /*
       * vchar lets us know our "limits".  Unadjust by the rotation amount,
       * so we can readjust by the rotation amount next time through.
       * (Could this be done easier?)
       */

      sp    = drawchar(loc, siznow, lab->angle, *text);
      loc.x = sp.x - rp.x * dl_aspect();
      loc.y = sp.y - rp.y;

      text++;
   }
}

/*===========================================================================
 *  dl_labelbounds(lab) -- return bounding box of given label
 *
 *  Return size in a DL_RECT, where the left edge of string is at the origin.
 *
 *         *---*-----------+
 *         *   *  ***  ****|
 *         ***** *   * *   *
 *         *   * *   * *   *
 *    -0-  *   *  ***  ****|
 *         |           *   |
 *         +-----------*---+
 *         0
 */


DL_RECT DX_API_PUBLIC dl_labelbounds(DL_LABEL *lab)
{
   DL_RECT box;  
   char    vstr[DM_ALPHALENGTH];  /* Char description string goes here. */
   float   width, above, below;   /* Width of string, distance above, below */
   float   siznow, offset;
   char    *text, *sizes, *super; /* local copies of string pointers */

   static float rsizes[10] = {
      1.0f, 0.9f, 0.8f, 0.7f, 0.6f, 0.5f, 0.4f, 0.3f, 0.2f, 0.1f
   };
   static float rsuper[10] = {
      0.0f, -0.8f, -0.6f, -0.4f, -0.2f, 0.0f, 0.2f, 0.4f, 0.6f, 0.8f
   };
    
   width = above = below = 0;	/* Zero the box dimensions. */

   offset= 0;			/* Set up default offset, */
   siznow= lab->size;		/* and the default character size. */

   text  = lab->text;
   sizes = lab->sizes;
   super = lab->super;
   while (*text) {
      if (*sizes) {
         if (isdigit(*sizes)) siznow = lab->size * rsizes[*sizes - '0'];
         sizes++;
      }
      if (*super) {
         if (isdigit(*super)) offset = lab->size * rsuper[*super - '0'];
         super++;
      }
	
      box    = vectorchar(*text, vstr);
      width += siznow * (box.xmax + 0.2f);
      if (offset + siznow * box.ymax > above)
         above = offset + siznow * box.ymax;
      if (offset + siznow * box.ymin < below)
         below = offset + siznow * box.ymin;

      text++;
   }

   width -= siznow * 0.2f;	/*** Subtract off .2 of last char. ***/

   /*** Package up answer, and return it. ***/

   box.xmin = 0;
   box.xmax = width;
   box.ymax = above;
   box.ymin = below;
   return box;
}

/*===========================================================================
 *  drawchar(loc, size, angle, c) -- draw a character, return sizes used
 */

static DL_POINT drawchar(DL_POINT loc, float size, float angle, char c)
{
   DL_POINT  point;
   DL_RECT   box;
   char      vstr[DM_ALPHALENGTH], *vp=vstr;

   /*** Get the description of how to "implement" c.  ***/

   box = vectorchar(c, vstr);

   /*** Now implement it. ***/

   while (*vp) {
      if (isupper(vp[0]))
         cmove(loc, 0.05f*size, -angle, tolower(vp[0]), vp[1]);
      else
         cdraw(loc, 0.05f*size, -angle, vp[0], vp[1]);
      vp += 2;
   }
    
   point.x = size * (box.xmax + 0.2f);
   point.y = 0.0f;

   /*** Figure out how much space this all took, and report it to caller. ***/

   return shiftpoint(loc, point, -angle);
}

/*===========================================================================
 *  shiftpoint() -- shift point by given amount (rotate, translate, scale)
 *
 *  This rotates offset by angle (in degrees), corrects for the current
 *  aspect ratio of the display, adds it to point, and returns the result.
 */

static DL_POINT shiftpoint(DL_POINT point, DL_POINT offset, float angle)
{
   DL_POINT p;
   float    theta, sint, cost;

   theta = DL_RADIAN(angle);

   if (fabs(theta) < 0.02) {
      sint = theta;
      cost = 1.0;
   } else {
      sint = (float)sin(theta);
      cost = (float)cos(theta);
   }
    
   p.x = point.x + (offset.x * cost + offset.y * sint) * dl_aspect();
   p.y = point.y + (offset.y * cost - offset.x * sint);

   return p;
}

/*===========================================================================
 *  rotatepoint(p, theta) -- rotate point about origin though angle th.
 */

static DL_POINT rotatepoint(DL_POINT point, float theta)
{
   DL_POINT rotated;
   float sint, cost;

   if (fabs(theta) < 0.02) {
      sint = theta;
      cost = 1.0f;
   } else {
      sint = (float)sin(theta);
      cost = (float)cos(theta);
   }
    
   rotated.x = point.x * cost + point.y * sint;
   rotated.y = point.y * cost - point.x * sint;
   return rotated;
}

/*===========================================================================
 *  vectorchar(c, vstr) -- return bounding box & description of char c
 */

static DL_RECT vectorchar(unsigned char c, char *vstr)
{
   DL_RECT box;

   if (! fontbuilt) font_init();

   strcpy(vstr, font[c]);
   box.xmin = 0.0f;
   box.xmax = 0.05f * wid[c];
   box.ymax = 0.05f * up[c];
   box.ymin = 0.05f * down[c];
   return box;
}


/*===========================================================================
 *  font_init() -- initialize vector font
 *
 *  Each character is represented as an (even length) string of characters.
 *  Each pair of characters in the string represents a vector end-point
 *  coordinate (in units of .05) based on its ASCII value.
 *  Upper case characters mean "move", lower case characters mean "draw".
 *  Odd characters represent height, digits being negative.
 */

static void font_init()
{
   int i;

   font[' ']="AaHa";
   font['A']="AacghqmgoaCgmg";
   font['B']="AaiakblcmelgkhiikjlkmmlokpiqaqBqbabiii";
   font['C']="Mclbjaeacbbcaeambocpeqjqlpmo";
   font['D']="AaiakblcmemmlokpiqaqBqba";
   font['E']="KqaqaakaAihi";
   font['F']="KqaqaaAihi";
   font['G']="MclbjaeacbbcaeambocpeqjqlpmoMamgHgog";
   font['H']="AaaqKakqAiki";
   font['I']="AagaAqgqDadq";
   font['J']="AgaebccbeagbhcieiqDqnq";
   font['K']="AaaqAgkqCika";
   font['L']="Kaaaaq";
   font['M']="Aaaqhgoqoa";
   font['N']="Aaaqmamq";
   font['O']="Aebccbeaiakblcmemmlokpiqeqcpboamae";
   font['P']="Aaaqiqkplommmlljkiihah";
   font['Q']="AebccbeaiakblcmemmlokpiqeqcpboamaeHfma";
   font['R']="AaaqiqkplommmlljkiihahIhma";
   font['S']="Adcbeaiakblcmelgkhcjbkambocpeqiqkpmn";
   font['T']="HahqAqoq";
   font['U']="Aqaebccbeaiakblcmemq";
   font['V']="Aqgamq";
   font['W']="Aqeahikaoq";
   font['X']="AqkaAakq";
   font['Y']="AqfikqFafi";
   font['Z']="Aqkqaaka";
   font['a']="BkdliljkkikblaKbiadabbadbfdgigke";
   font['b']="BccbeagaibjckekgjiijgkekcjbiAqbqbaaa";
   font['c']="Jkhlelckbjahaebccbeahajb";
   font['d']="KcgaeacbbcaeahbjckelhljkkiLakakqlq";
   font['e']="Kcibgaeacbbcaeahbjckelhljkkikgag";
   font['f']="BadaCacldnfohnilAhfh";
   font['g']="KcibgaeacbbcaeagbidjhjjikgLjkjk3j5h6e6c5b4";
   font['h']="AqbqbaaaBhcjekhkjjkhka";
   font['i']="AaeaCacjbjClcmbmblcl";
   font['j']="GmhmhlglgmGjhjh2g4e5d5b4a2";
   font['k']="AababqaqBdhjIaeg";
   font['l']="AaeaCacobo";
   font['m']="AkbkbaBhcjekfkhjihiaihjjlkmkojphpa";
   font['n']="AkbkbaaaBhcjekhkjjkhka";
   font['o']="Aebccbeagaibjckekgjiijgkekcjbiagae";
   font['p']="BccbeagaibjckekgjiijgkekcjbiAkbkb6a6";
   font['q']="KcgaeacbbcaeahbjckelhljkkiLjkjk3l5n6";
   font['r']="AkbkbaBhcjekhkji";
   font['s']="Abcagaibjdiebgahbjdkhkjj";
   font['t']="AkikEoebfahaib";
   font['u']="AkaebccbeagaibjckeKkkala";
   font['v']="Alfakl";
   font['w']="Aleahgkaol";
   font['x']="AlkaAakl";
   font['y']="AkaebccbeagaibjckeKkk3j5h6e6c5b4";
   font['z']="Alklaaka";
   font['0']="DqbpanadbbdahajbkdknjphqdqBbjp";
   font['1']="AndqdaAaga";
   font['2']="Kaaaacbehijjklknjphqdqbpan";
   font['3']="AnbpdqhqjpknkljjhieiHijhkfkdjbhadabbad";
   font['4']="Kfafhqha";
   font['5']="Kqaqaicjhjjikgkdjbhacaab";
   font['6']="Jphqdqbpanadbbdahajbkdkgjihjdjbiah";
   font['7']="Aqkqda";
   font['8']="DqbpanalbjdihijjklknjphqdqDibhafadbbdahajbkdkfjhhi";
   font['9']="Bbdahajbkdknjphqdqbpanakbidhhhjikk";
   font['!']="CbdbdccccbCedeeqbqce";
   font['#']="AfmfAkmkDcdnJcjn";
   font['^']="Akeqik";
   font['&']="Kaibbmbncpeqfqgogmfkbiagaebccbeafahbke";
   font['$']="AdcchcjdkfjhhidibjalbndoioknFafq";
   font['\'']="Epdpdqeqeocm";
   font['`']="Cpdpdqcqcoem";
   font['"']="CpbpbqcqcobnFpepeqfqfoen";
   font['.']="Dacacbdbda";
   font[',']="Dacacbdbdad1c2";
   font[':']="DccccddddcCjdjdkckcj";
   font[';']="DccccddddcdbcaCjdjdkckcj";
   font['~']="Albncoeofnfmgliljmko";
   font['_']="Aaka";
   font['-']="Aiii";
   font['+']="AiiiEeem";
   font['>']="Aakiaq";
   font['<']="Kaaikq";
   font['(']="Eacbbcaeambocpeq";
   font[')']="Aacbdceeemdocpaq";
   font[']']="Aafafqaq";
   font['[']="Faaaaqfq";
   font['{']="Eacbbdbfchaicjblbncpeq";
   font['}']="Aacbdddfcheicjdldncpaq";
   font['?']="AlambocpeqgqipjoknkljjhighfffdFbfagagbfb";
   font['=']="AmimAhjh";
   font['%']="AamqIblbmcmflgighfhcibCkfkglgofpcpboblck";
   font['/']="Aamq";
   font['\\']="Aqma";
   font['|']="Aaaq";
   font['@']="Khjfheffehejflhmjlkkkeldmdneogomnompkqeqcpboamaebccbeaka";
   font['*']="AjljCejoCoje";

   for (i=0; i<DM_ALPHACOUNT; i++) {
      char *s;
      int  x, y;
      dt_Real v;

      wid[i] = up[i] = down[i] = 0;

      s = font[i];
      if (! s) continue;
      while (*s) {
         x = s[0];
         y = s[1];
      
         if (charvalue(x) > wid[i]) wid[i] = charvalue(x);
         v = charvalue(y);
         if (v > up[i]) up[i] = v;
         if (v < down[i]) down[i] = v;
      
         s += 2;
      }
   }
    
   fontbuilt++;
}

/*===========================================================================
 *  cmove() -- move in character coordinates
 */

static void cmove(DL_POINT loc, float size, float angle, char cx, char cy)
{
    DL_POINT p;	      /* Point indicated by cx, cy. */

    p.x  = size * charvalue(cx);
    p.y  = size * charvalue(cy);

    p = shiftpoint(loc, p, angle);
    dl_moveto(&p);
}

/*===========================================================================
 *  cdraw(c) -- draw in character coordinates
 */

static void cdraw(DL_POINT loc, float size, float angle, int cx, int cy)
{
    DL_POINT p;	       /* Point indicated by cx, cy. */

    p.x = size * charvalue(cx);
    p.y = size * charvalue(cy);

    p = shiftpoint(loc, p, angle);
    dl_drawto(&p);
}

/*===========================================================================
 *  charvalue(c) -- return coded character location value
 */

static dt_Real charvalue(char c)
{
   if (isdigit(c))
      return (dt_Real)('0' - c); /* Negative values for non-zero digits. */
   if (isupper(c))               /* Alphabet positions for others. */
      return (dt_Real)(c - 'A');
   return (dt_Real)(c - 'a');
}

float dl_aspect(void)
{ return 1.0; }
